from collections import namedtuple
from unittest.mock import MagicMock

import pytest

from infection_monkey.exploit import powershell
from infection_monkey.exploit.consts import WIN_ARCH_32, WIN_ARCH_64
from infection_monkey.exploit.powershell_utils.auth_options import AuthOptions
from infection_monkey.exploit.powershell_utils.credentials import Credentials
from infection_monkey.model.host import VictimHost

# Use the path_win32api_get_user_name fixture for all tests in this module
pytestmark = pytest.mark.usefixtures("patch_win32api_get_user_name")

USER_LIST = ["user1", "user2"]
PASSWORD_LIST = ["pass1", "pass2"]
LM_HASH_LIST = ["bogo_lm_1"]
NT_HASH_LIST = ["bogo_nt_1", "bogo_nt_2"]
DROPPER_TARGET_PATH_32 = "C:\\agent32"
DROPPER_TARGET_PATH_64 = "C:\\agent64"

Config = namedtuple(
    "Config",
    [
        "exploit_user_list",
        "exploit_password_list",
        "exploit_lm_hash_list",
        "exploit_ntlm_hash_list",
        "dropper_target_path_win_32",
        "dropper_target_path_win_64",
    ],
)


class TestAuthenticationError(Exception):
    pass


@pytest.fixture
def powershell_exploiter(monkeypatch):
    host = VictimHost("127.0.0.1")
    pe = powershell.PowerShellExploiter(host)
    pe._config = Config(
        USER_LIST,
        PASSWORD_LIST,
        LM_HASH_LIST,
        NT_HASH_LIST,
        DROPPER_TARGET_PATH_32,
        DROPPER_TARGET_PATH_64,
    )

    monkeypatch.setattr(powershell, "AuthenticationError", TestAuthenticationError)
    monkeypatch.setattr(powershell, "is_windows_os", lambda: True)
    # It's regrettable to mock out a private method on the PowerShellExploiter instance object, but
    # it's necessary to avoid having to deal with the monkeyfs
    monkeypatch.setattr(pe, "_write_virtual_file_to_local_path", lambda: None)

    return pe


def test_powershell_disabled(monkeypatch, powershell_exploiter):
    mock_powershell_client = MagicMock(side_effect=Exception)
    monkeypatch.setattr(powershell, "PowerShellClient", mock_powershell_client)

    success = powershell_exploiter.exploit_host()
    assert not success


def test_powershell_http(monkeypatch, powershell_exploiter):
    def allow_http(_, credentials: Credentials, auth_options: AuthOptions):
        if not auth_options.ssl:
            raise TestAuthenticationError
        else:
            raise Exception

    mock_powershell_client = MagicMock(side_effect=allow_http)
    monkeypatch.setattr(powershell, "PowerShellClient", mock_powershell_client)
    powershell_exploiter.exploit_host()

    for call_args in mock_powershell_client.call_args_list:
        assert not call_args[0][2].ssl


def test_powershell_https(monkeypatch, powershell_exploiter):
    def allow_https(_, credentials: Credentials, auth_options: AuthOptions):
        if auth_options.ssl:
            raise TestAuthenticationError
        else:
            raise Exception

    mock_powershell_client = MagicMock(side_effect=allow_https)
    monkeypatch.setattr(powershell, "PowerShellClient", mock_powershell_client)
    powershell_exploiter.exploit_host()

    for call_args in mock_powershell_client.call_args_list:
        if call_args[0][1].secret != "" and call_args[0][1].secret != "dummy_password":
            assert call_args[0][2].ssl


def test_no_valid_credentials(monkeypatch, powershell_exploiter):
    mock_powershell_client = MagicMock(side_effect=TestAuthenticationError)
    monkeypatch.setattr(powershell, "PowerShellClient", mock_powershell_client)

    success = powershell_exploiter.exploit_host()
    assert not success


def authenticate(mock_client):
    def inner(_, credentials: Credentials, auth_options: AuthOptions):
        if credentials.username == "user1" and credentials.secret == "pass2":
            return mock_client
        else:
            raise TestAuthenticationError("Invalid credentials")

    return inner


@pytest.mark.parametrize(
    "dropper_target_path,arch",
    [(DROPPER_TARGET_PATH_32, WIN_ARCH_32), (DROPPER_TARGET_PATH_64, WIN_ARCH_64)],
)
def test_successful_copy(monkeypatch, powershell_exploiter, dropper_target_path, arch):
    mock_client = MagicMock()
    mock_client.return_value.get_host_architecture = lambda: arch
    mock_client.return_value.copy_file = MagicMock(return_value=True)

    monkeypatch.setattr(powershell, "PowerShellClient", mock_client)

    success = powershell_exploiter.exploit_host()

    assert dropper_target_path in mock_client.return_value.copy_file.call_args[0][1]
    assert success


def test_failed_copy(monkeypatch, powershell_exploiter):
    mock_client = MagicMock()
    mock_client.return_value.get_host_architecture = lambda: WIN_ARCH_32
    mock_client.return_value.copy_file = MagicMock(return_value=False)

    monkeypatch.setattr(powershell, "PowerShellClient", mock_client)

    success = powershell_exploiter.exploit_host()
    assert not success


def test_failed_monkey_execution(monkeypatch, powershell_exploiter):
    mock_client = MagicMock()
    mock_client.get_host_architecture = lambda: WIN_ARCH_32
    mock_client.copy_file = MagicMock(return_value=True)
    mock_client.execute_cmd_as_detached_process = MagicMock(side_effect=Exception)

    mock_powershell_client = MagicMock(side_effect=authenticate(mock_client))
    monkeypatch.setattr(powershell, "PowerShellClient", mock_powershell_client)

    success = powershell_exploiter.exploit_host()
    assert not success


def test_login_attemps_correctly_reported(monkeypatch, powershell_exploiter):
    mock_client = MagicMock()
    mock_client.return_value.get_host_architecture = lambda: WIN_ARCH_32
    mock_client.return_value.copy_file = MagicMock(return_value=True)

    # execute_cmd method will throw exceptions for 5 first calls.
    # 6-th call doesn't throw an exception == credentials successful
    execute_cmd_returns = [Exception, Exception, Exception, Exception, Exception, True]
    mock_client.return_value.execute_cmd = MagicMock(side_effect=execute_cmd_returns)

    monkeypatch.setattr(powershell, "PowerShellClient", mock_client)

    powershell_exploiter.exploit_host()

    # Total 6 attempts reported, 5 failed and 1 succeeded
    assert len(powershell_exploiter.exploit_attempts) == len(execute_cmd_returns)
    assert (
        len([attempt for attempt in powershell_exploiter.exploit_attempts if not attempt["result"]])
        == 5
    )
    assert (
        len([attempt for attempt in powershell_exploiter.exploit_attempts if attempt["result"]])
        == 1
    )


def test_build_monkey_execution_command():
    host = VictimHost("127.0.0.1")
    depth = 2
    executable_path = "/tmp/test-monkey"

    cmd = powershell.build_monkey_execution_command(host, depth, executable_path)

    assert f"-d {depth}" in cmd
    assert executable_path in cmd
